<!doctype html>
<html lang="en">

<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{{ title }}</title>
    <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.3/dist/css/bootstrap.min.css" rel="stylesheet"
        integrity="sha384-QWTKZyjpPEjISv5WaRU9OFeRpok6YctnYmDr5pNlyT2bRjXh0JMhjY6hW+ALEwIH" crossorigin="anonymous">
</head>

<body>
    <nav class="navbar navbar-expand-lg navbar-light bg-light">
        <div class="container-fluid">
            <a href="#shapes">List of shapes</a>
            <a href="#uml">Visualization</a>
            <a href="#metadata">Metadata</a>
        </div>
    </nav>

    <h1>
        Schema: {{ title }}
    </h1>
    <div class="container">
        <h2 id="shapes">List of shapes</h2>
        <table class="table table-striped">
            <thead>
                <tr>
                    <th>Shape ID</th>
                    <th>Shape Label</th>
                </tr>
            </thead>

            {% for shape in shapes %}
            <tr>
                <td>
                    <a href="{{ shape.href }}">{{ shape.name }}</a>
                </td>
                <td>
                    {{ shape.label }}
                </td>
            </tr>
            {% endfor %}
        </table>
    </div>
    <h2 id="uml">UML-based diagram</h2>
    <figure class="figure">
        {% autoescape false %}{{ svg_schema }}{% endautoescape %}
        <figcaption class="figure-caption text-end">Generated UML-like diagram of the ShEx schema</figcaption>
    </figure>

    <h2 id="metadata">Metadata</h2>
    <ul>
        <li>rudof_version: {{rudof_version}}</li>
        <li>created time: {{ created_time }}</li>
    </ul>
    <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.3/dist/js/bootstrap.bundle.min.js"
        integrity="sha384-YvpcrYf0tY3lHB60NNkmXc5s9fDVZLESaAA55NDzOxhy9GkcIdslK1eN7N6jIeHz"
        crossorigin="anonymous"></script>
</body>

</html>
