pub mod arrow_style;
pub mod rdf_visualizer_config;
pub mod rdf_visualizer_error;
pub mod stereotype_style;
pub mod style;
pub mod thickness_style;
pub mod uml_color;
pub mod usage_count;
pub mod visual_rdf_edge;
pub mod visual_rdf_graph;
pub mod visual_rdf_node;

const REIFIES: &str = "http://www.w3.org/1999/02/22-rdf-syntax-ns#reifies";
