//! `rudof_lib` presents the public API to interact with `rudof` programmatically
//!
//!
#![deny(rust_2018_idioms)]
pub mod data;
pub mod data_format;
pub mod input_spec;
pub mod node_info;
pub mod query;
pub mod query_result_format;
pub mod query_type;
pub mod result_shacl_validation_format;
pub mod result_shex_validation_format;
pub mod rudof;
pub mod rudof_config;
pub mod rudof_error;
pub mod selector;
pub mod shacl;
pub mod shacl_format;
pub mod shapemap_format;
pub mod shapes_graph_source;
pub mod shex;
pub mod shex_format;
pub mod sort_by_result_shape_map;
pub mod terminal_width;
#[cfg(target_family = "wasm")]
mod wasm_stubs;

pub use input_spec::*;
pub use iri_s::*;
pub use oxrdf;
pub use rudof::*;
pub use rudof_config::*;
pub use rudof_error::*;
pub use selector::*;
pub use shacl_ir;
pub use shacl_validation;
pub use shapes_graph_source::*;
pub use shex_ast::*;
pub use srdf;
