// Current modules

#[cfg(target_family = "wasm")]
compile_error!("This crate is not intended to be used in a WebAssembly environment.");

pub mod cli;
pub mod color_support;
pub mod compare;
pub mod convert;
pub mod data;
pub mod dctap;
pub mod dctap_format;
pub mod dctap_result_format;
pub mod generate_schema_format;
pub mod input_compare_format;
pub mod input_compare_mode;
pub mod input_convert_format;
pub mod input_convert_mode;
pub mod node;
pub mod output_convert_format;
pub mod output_convert_mode;
pub mod pgschema;
pub mod query;
pub mod rdf_config;
pub mod rdf_reader_mode;
pub mod result_compare_format;
pub mod result_data_format;
pub mod result_service_format;
pub mod result_validation_format;
pub mod service;
pub mod shacl;
pub mod shapemap;
pub mod shex;
pub mod show_mode;
pub mod sort_by_validate;
pub mod validation_mode;
pub mod writer;

pub use color_support::*;
pub use compare::*;
pub use convert::*;
pub use dctap::*;
pub use dctap_result_format::*;
pub use generate_schema_format::*;
pub use input_compare_format::*;
pub use input_compare_mode::*;
pub use input_convert_format::*;
pub use input_convert_mode::*;
pub use output_convert_format::*;
pub use output_convert_mode::*;
pub use pgschema::*;
pub use rdf_config::*;
pub use rdf_reader_mode::*;
pub use result_data_format::*;
pub use result_service_format::*;
pub use result_validation_format::*;
pub use service::*;
pub use shacl::*;
pub use shapemap::*;
pub use shex::*;
pub use show_mode::*;
pub use sort_by_validate::*;
pub use validation_mode::*;
