//! Test suite for the Rudof MCP server.
//!
//! This module contains integration tests for:
//! - Server capabilities (tools, prompts, resources)
//! - Service initialization and handlers

mod capabilities_tests;
mod service_tests;
